/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.List;

import org.sablecc.sablecc.analysis.DepthFirstAdapter;
import org.sablecc.sablecc.node.*;

public class GenTokens extends DepthFirstAdapter
{
  private MacroExpander macros;
  private ResolveIds ids;
  private File pkgDir;

  //    final GenTokens instance = this;
  String text;

  public GenTokens(ResolveIds ids)
  {
    this.ids = ids;

    try
    {
      macros = new MacroExpander(
                 new InputStreamReader(
                   getClass().getResourceAsStream("tokens.txt"), StandardCharsets.UTF_8));
    }
    catch(IOException e)
    {
      throw new RuntimeException("unable to open tokens.txt.", e);
    }

    pkgDir = new File(ids.pkgDir, "node");

    if(!pkgDir.exists())
    {
      if(!pkgDir.mkdir())
      {
        throw new RuntimeException("Unable to create " + pkgDir.getAbsolutePath());
      }
    }
  }

  @Override
  public void inATokenDef(ATokenDef node)
  {
    String name = ids.names.get(node);
    text = null;

    ARegExp regExp = (ARegExp) node.getRegExp();

    List<PConcat> concats = regExp.getConcats();

    if(concats.size() == 1)
    {
      AConcat concat = (AConcat)concats.get(0);
      List<PUnExp> unExps = concat.getUnExps();

      if(unExps.size() == 1)
      {
        AUnExp unExp = (AUnExp) unExps.get(0);

        PBasic basic = unExp.getBasic();

        if((basic instanceof AStringBasic) &&
            (unExp.getUnOp() == null))
        {
          text = ((AStringBasic) basic).getString().getText();
          text = text.substring(1, text.length() - 1);
        }
        else if((basic instanceof ACharBasic) &&
                (unExp.getUnOp() == null))
        {
          PChar pChar = ((ACharBasic) basic).getChar();

          if(pChar instanceof ACharChar)
          {
            text = ((ACharChar) pChar).getChar().getText();
            text = text.substring(1, text.length() - 1);
          }
        }
      }
    }

    try(BufferedWriter file = Files.newBufferedWriter(new File(pkgDir, name + ".java").toPath(), StandardCharsets.UTF_8))
    {
      if(text == null)
      {
        ids.fixedTokens.put(node, false);

        macros.apply(file, "VariableTextToken", new String[] {ids.pkgNameDot, name});
      }
      else
      {
        ids.fixedTokens.put(node, true);

        macros.apply(file, "FixedTextToken", new String[] {ids.pkgNameDot, name, processText(text)});

        ids.errorNames.put(node, "'" + text + "'");
      }
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, name + ".java").getAbsolutePath(), e);
    }
  }

  private String processText(String s)
  {
    StringBuilder result = new StringBuilder();

    for(int i = 0; i < s.length(); i++)
    {
      char c = s.charAt(i);

      switch(c)
      {
      case '\b':
        {
          result.append("\\t");
          break;
        }
      case '\t':
        {
          result.append("\\t");
          break;
        }
      case '\n':
        {
          result.append("\\n");
          break;
        }
      case '\f':
        {
          result.append("\\f");
          break;
        }
      case '\r':
        {
          result.append("\\r");
          break;
        }
      case '\"':
        {
          result.append("\\\"");
          break;
        }
      case '\'':
        {
          result.append("\\\'");
          break;
        }
      case '\\':
        {
          result.append("\\\\");
          break;
        }
      default:
        {
          result.append(c);
        }
      }
    }

    return result.toString();
  }
}

